      *_> CNLLSTSPLF SRCFILE(@2/@1) SRCMBR(@3)
      *_> DLTMOD MODULE(@5/@4)
      *_> CRTRPGMOD MODULE(@5/@4) SRCFILE(@2/@1) SRCMBR(@3) +
      *_>   DBGVIEW(@9) OPTIMIZE(@8) OPTION(*EVENTF)
      * ---------------------------------------------------------------
      * Module.: TG_MED_m02           Project.....:
      * Author.: A. Campin            Date written: 06/04/2008
      * Purpose: Trigger Mediator. Function MediateTriggerCall is called by "C"
      *            with the trigger buffer. This function mediates between the
      *            database manager and the actual service program doing the work.
      *            Its only function is to call the approriate service program
      *            for that file. All other functionality is contained in
      *            service programs. This is done so that the only lock
      *            maintained on a program is on this program which is as
      *            simple as possible. If there is an error in a trigger
      *            program, we can repair it without having to bring
      *            down the entire system to remove the locks.
      *          All trigger service programs are loaded dynamically the
      *            first time they are needed. If a data base table is
      *            never opened up for update, the service program
      *            associated with that data base table will never be
      *            loaded!
      * ---------------------------------------------------------------
      * Called by: OS/400 Data Base.
      * ---------------------------------------------------------------
      * Revision history:
      *  Project # Pgmr        Date   Desc
      *
      * End Revision History
      * ---------------------------------------------------------------
     h NoMain
     h Option(*Srcstmt:*Nodebugio)

      /copy *libl/qsrcf,cb_StdType
      /copy *libl/qsrcf,cb_Std_Con

      /copy *libl/qsrcf,tg_Med_pr

      /copy *libl/qsrcf,xverrh_pr
      /copy *libl/qsrcf,xvuspc_pr
      /copy *libl/qsrcf,xvdynl_pr

     d MediateTriggerCall...
     d                 pr                  ExtProc('MediateTriggerCall')
     d   PR_InPtrTriggerBuffer...
     d                                     Like(StdPtr)
     d                                     Value

     d BSearchCompare...
     d                 pr                  Like(StdInt)
     d                                     ExtProc('BSearchCompare')
     d   PR_PointerToLookForValue...
     d                                     Like(StdPtr)
     d                                     Value
     d   PR_PointerToArrayElement...
     d                                     Like(StdPtr)
     d                                     Value

      *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      * Data Structure for Trigger Control Header. Mapped to User Space
      *  TG_CTRL_HD.
     d dsTriggerControlHeader...
     d               e ds                  Qualified
     d                                     Extname(TGCTRLH)
     d                                     Based(ptrTriggerControlHeader)

     d g_PtrBaseOfDetailRecords...
     d                 s                   Like(StdPtr)
     d                                     Import
     d g_NumberOfDetailRecordsLoaded...
     d                 s                   Like(StdInt)
     d                                     Import
      * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      * MediateTriggerCall
      *  This function is called by 'C' program and does actual mediation
      *    between our program and service program handling the actual work.
      *   Input       - Pointer To Trigger Buffer.
      *   Out         - None
      *   Returns     - None
      * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     p MediateTriggerCall...
     p                 b                   Export
     d                 pi
     d   InPtrTriggerBuffer...
     d                                     Like(StdPtr)
     d                                     Value

     d FindName...
     d                 pr                  Like(StdPtr)
     d                                     ExtProc('bsearch')
     d   PR_PtrLookForValue...
     d                                     Like(StdPtr)
     d                                     Value
     d   PR_PtrArrayElement...
     d                                     Like(StdPtr)
     d                                     Value
     d   PR_NumberOfElements...
     d                                     Like(StdIntUns)
     d                                     Value
     d   PR_SizeOfEachElement...
     d                                     Like(StdIntUns)
     d                                     Value
     d   PR_PtrCompareProcedure...
     d                                     Like(StdPrcPtr)
     d                                     Value

     d cControlUserSpaceName...
     d                 c                   Const('TG_CTRL_HD')
     d cControlUserSpaceLibrary...
     d                 c                   Const('*LIBL')

     d pptrCallProcedure...
     d                 s                   Like(StdPrcPtr)
     d CallProcedure...
     d                 pr                  ExtProc(pptrCallProcedure)
     d   PR_InPtrTriggerBuffer...
     d                                     Like(StdPtr)
     d                                     Const

      *- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      * Structure of Detail Record available for triggers.
     d dsDetailRecord...
     d                 ds                  LikeDs(TD_DetailRecord)
     d                                     Based(ptrDetailRecord)

     d dsQualifiedFileName...
     d                 ds                  Qualified
     d  FileName...
     d                                     Like(StdNam)
     d  LibraryName...
     d                                     Like(StdNam)

      /copy *libl/qsrcf,cb_dbtrg_s

     d LastReloadValue...
     d                 s                   Like(StdTMS)
     d                                     Inz(*LoVal)
     d                                     Static
     d ActivationMark...
     d                 s                   Like(StdInt)
      /Free

       If ptrTriggerControlHeader = *Null;
         ptrTriggerControlHeader =
            USPC_GetPointerToUserSpace(cControlUserSpaceName   :
                                       cControlUserSpaceLibrary);
       EndIf;

       ptrDbTriggerBuffer = InPtrTriggerBuffer;

       // If Triggers not active, just return.
       If dsTriggerControlHeader.THACTIVE = cNo;
         Return;
       EndIf;

       // If reload timestamp has from last time program called or
       // original load, reload detail tables.
       If LastReloadValue <> dsTriggerControlHeader.THRELOADTS;
         LastReloadValue = dsTriggerControlHeader.THRELOADTS;
         LoadDetailRecords();
       EndIf;

       // If not found, then not active for triggers.
       If g_NumberOfDetailRecordsLoaded <= 0;
         Return;
       EndIf;

       // See if file is active for triggers.
       dsQualifiedFileName.FileName    = DbTriggerBuffer.PhysicalFileName;
       dsQualifiedFileName.LibraryName = DbTriggerBuffer.PhysicalFileLibrary;
       ptrDetailRecord = FindName(%Addr(dsQualifiedFileName)   :
                                  g_PtrBaseOfDetailRecords     :
                                  g_NumberOfDetailRecordsLoaded:
                                  %Size(dsDetailRecord)        :
                                  %PAddr('BSearchCompare')     );
       If ptrDetailRecord = *Null;
         Return;
       EndIf;

       // If not loaded before, load the service program.
       If dsDetailRecord.pptrTriggerProcedure = *Null;
         ActivationMark =
           DYNL_ActivateServiceProgram(dsDetailRecord.ServiceProgramName   :
                                       dsDetailRecord.ServiceProgramLibrary);
         dsDetailRecord.pptrTriggerProcedure =
           DYNL_GetPointerToExportedProcedure(ActivationMark                 :
                                              dsDetailRecord.ServiceProcedure);
       EndIf;

       // Call the procedure to handle this file.
       pptrCallProcedure = dsDetailRecord.pptrTriggerProcedure;
       Monitor;
         CallProcedure(InPtrTriggerBuffer);
       On-Error;
         ERRH_Throw('CPF9898'                                               :
                    cCPFMessageFileName                                     :
                    'Call to service program ' +
                       %TrimR(dsDetailRecord.ServiceProgramName) + ' failed');
       EndMon;

       Return;

      /End-Free
     p                 e
      * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      * BSearchCompare
      *   Called by bsearch to compare two values and return result.
      *   Input     - Pointer to Look for value.
      *               Pointer to Array Element.
      *   Out       - None
      *   Returns   - Result Code.
      * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     p BSearchCompare...
     p                 b
     d                 pi                  Like(StdInt)
     d    PointerToLookForValue...
     d                                     Like(StdPtr)
     d                                     Value
     d    PointerToArrayElement...
     d                                     Like(StdPtr)
     d                                     Value

     d LookForKey...
     d                 s                   Like(StdQNam)
     d                                     Based(PointerToLookForValue)
     d ArrayElement...
     d                 s                   Like(StdQNam)
     d                                     Based(PointerToArrayElement)

     d cLessThan...
     d                 c                   Const(-1)
     d cGreaterThan...
     d                 c                   Const(1)
     d cEqualTo...
     d                 c                   Const(0)
      /Free

          Select;
            When LookForKey < ArrayElement;
              Return cLessThan;
            When LookForKey = ArrayElement;
              Return cEqualTo;
            When LookForKey > ArrayElement;
              Return cGreaterThan;
          EndSl;

      /End-Free
     p                 e
