      *_> CNLLSTSPLF SRCFILE(@2/@1) SRCMBR(@3)
      *_> DLTMOD MODULE(@5/@4)
      *_> CRTRPGMOD MODULE(@5/@4) SRCFILE(@2/@1) SRCMBR(@3) +
      *_>   DBGVIEW(@9) OPTIMIZE(@8)
      *=============================================================================================
      * Module.....: XVERRH_M02
      * By.........: Alan Campin
      * Date.......: 06/13/2008
      * Purpose....: Error Handling Functions. This module contains Message
      *                Storage functions used by error handling.
      *                Each time an error occurs, functions call this module
      *                to store the message. A call to ERRH_GetLastMessage
      *                will retreive the last error message. A call to
      *                ERRH_GetMessage will retrieve first message and remove from
      *                list.
      *
      *              Note: These functions are implemented using a Linked
      *                      List. Each item allocated contains one
      *                      message with each message pointing to the next
      *                      message except the last entry which is always
      *                      Null. New messages are added to the
      *                      end of the list and when being read out, the
      *                      queue is implemented as a FIFO (First In,
      *                      First Out) list so the first message is
      *                      pulled out and Anchor pointed to the next
      *                      item in the list until all have been read
      *                      out.
      *---------------------------------------------------------------------------------------------
      * Program Modification Section:
      *------------------------------
      * By.........:
      * Date.......:
      * Description:                                                        l the
      *            :
      *
      *=============================================================================================
     h nomain

      /copy *libl/qsrcf,cb_StdType
      /copy *libl/qsrcf,cb_Std_Con
      /copy *libl/qsrcf,xverrh_pr

     d AddMessageToList...
     d                 pr
     d   PR_InMessageId...
     d                                     Like(StdMsgId)
     d                                     Value
     d   PR_InQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d                                     Value
     d   PR_InMessageData...
     d                             1024    Varying
     d                                     Value

     d GetMessageFromList...
     d                 pr                  Like(StdLgl)
     d   PR_InPointerToEntry...
     d                                     Like(StdPtr)
     d                                     Value
     d   PR_OutMessageId...
     d                                     Like(StdMsgId)
     d   PR_OutQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d   PR_OutMessageData...
     d                             1024    Varying

     d RemoveMessageFromList...
     d                 pr
     d   PR_InPointerToEntry...
     d                                     Like(StdPtr)
     d                                     Value
     * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     * Global Variables
     d ptrAnchor...
     d                 s                   Like(StdPtr)
     d                                     Inz(*Null)
     d ptrTail...
     d                 s                   Like(StdPtr)
     d                                     Inz(*Null)
     d MessageCount...
     d                 s                   Like(StdIntSml)

     d MessageList...
     d                 ds                  Based(ptrMessageList)
     d                                     Qualified
     d  MessageId...
     d                                     Like(StdMsgId)
     d  QualifiedMessageFileName...
     d                                     Like(StdQNam)
     d  MessageDataLength...
     d                                     like(StdIntSml)
     d  PointerToNextEntry...
     d                                     Like(StdPtr)

     d MessageData...
     d                 s           1024    Based(ptrMessageData)
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * ERRH_AddMessage
     *   Adds a message to storage.
     *       In     Parms - Msgid of Error.
     *                      Qualified Message File Name.
     *                      Message Data.
     *       In/Out Parms - None.
     *       Out    Parms - None.
     *     Return         - None.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p ERRH_AddMessage...
     p                 b                   Export
     d                 pi
     d   InMessageId...
     d                                     Like(StdMsgId)
     d                                     Value
     d   InQualifiedMessageFileName...
     d                                     Like(StdQNam)
     d                                     Value
     d   InMessageData...
     d                             1024    Varying
     d                                     Value
      /Free

       AddMessageToList(InMessageId               :
                        InQualifiedMessageFileName:
                        InMessageData             );
       Return;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * ERRH_GetMessageCount
     *   Get count of number of messages written to storage.
     *       In     Parms - None.
     *       In/Out Parms - None.
     *       Out    Parms - None.
     *     Return         - Count of Messages in storage.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p ERRH_GetMessageCount...
     p                 b                   Export
     d                 pi                  Like(StdIntSml)

      /Free

       Return MessageCount;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * ERRH_GetLastMessage.
     *   This function just gets the last message written but does it
     *     without releasing storage.
     *       In     Parms - None.
     *       In/Out Parms - None.
     *       Out    Parms - Msg ID.
     *                      Qualified Message File Name.
     *                      Message Data.
     *     Return         - True or False-Message Found.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p ERRH_GetLastMessage...
     p                 b                   Export
     d                 pi                  Like(StdLgl)
     d   OutMessageId...
     d                                     Like(StdMsgId)
     d   OutQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d   OutMessageData...
     d                             1024    Varying
      /Free

       //Set return value to default.
       OutMessageId                = *Blanks;
       OutQualifiedMessageFileName = *Blanks;
       %Len(OutMessageData)        = *Zero;

       //If empty list, just return False to found.
       If ptrTail = *Null;
         Return cFalse;
       EndIf;

       Return GetMessageFromList(ptrTail                    :
                                 OutMessageId               :
                                 OutQualifiedMessageFileName:
                                 OutMessageData             );
      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * ERRH_GetMessage
     *   This function get the first message from the list and releases
     *     storage.
     *       In     Parms - None.
     *       In/Out Parms - None.
     *       Out    Parms - MessageId.
     *                      Qualified Message File Name.
     *                      Message Data.
     *     Return         - True or False. Message Found.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p ERRH_GetMessage...
     p                 b                   Export
     d                 pi                  Like(StdLgl)
     d   OutMessageId...
     d                                     Like(StdMsgId)
     d   OutQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d   OutMessageData...
     d                             1024    Varying
      /Free

       //Set return value to default.
       OutMessageId                = *Blanks;
       OutQualifiedMessageFileName = *Blanks;
       %Len(OutMessageData)        = *Zero;

       //If empty list, return false to message found.
       If ptrAnchor = *Null;
         Return cFalse;
       EndIf;

       //Get first entry in list. If found, we need to remove from list.
       If GetMessageFromList(ptrAnchor                  :
                             OutMessageId               :
                             OutQualifiedMessageFileName:
                             OutMessageData             );
         RemoveMessageFromList(ptrAnchor);
         Return cTrue;
       Else;
         Return cFalse;
       EndIf;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * ERRH_FreeAllMessages
     *   This function just releases all messages. Called if user does not
     *     want to read back message.
     *       In     Parms - None.
     *       In/Out Parms - None.
     *       Out    Parms - None.
     *   Returns          - None.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p ERRH_FreeAllMessages...
     p                 b                   Export
     d                 pi

     d ptrSaveNext...
     d                 s                   Like(StdPtr)
      /Free

       //If empty list, just return.
       If ptrTail = *Null;
         Return;
       EndIf;

       //Loop through messages and release.
       ptrMessageList = ptrAnchor;
       DoU 0 = 1;
         ptrSaveNext = MessageList.PointerToNextEntry;

         RemoveMessageFromList(ptrMessageList);

         If ptrSaveNext = *Null;
           Leave;
         EndIf;

         ptrMessageList = ptrSaveNext;
       EndDo;

       Return;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * AddMessageToList
     *   Add a message to list.
     *     In     Parms - MessageId.
     *                    Qualified Message File Name.
     *                    Message Data.
     *     In/Out Parms - None.
     *     Out    Parms - None.
     *   Return         - None.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p AddMessageToList...
     p                 b
     d                 pi
     d   InMessageId...
     d                                     Like(StdMsgId)
     d                                     Value
     d   InQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d                                     Value
     d   InMessageData...
     d                             1024    Varying
     d                                     Value

     d ptrTemp...
     d                 s                   Like(StdPtr)
      /Free

       //Allocate enough storage for header plus message. We do this
       //  so that we only have to one allocate operation.
       ptrTemp = %Alloc(%Size(MessageList) + %Len(InMessageData));

       //If this is first time, just set the anchor to point the first
       // entry. Otherwise, get the previous entry and point that next
       // entry to the entry we just created.
       If ptrAnchor = *Null;
         ptrAnchor = ptrTemp;
       Else;
         ptrMessageList = ptrTail;
         MessageList.PointerToNextEntry = ptrTemp;
       EndIf;

       //Point to current entry.
       ptrTail        = ptrTemp;
       ptrMessageList = ptrTemp;

       //Move in Message data fields.
       MessageList.MessageID                = InMessageId;
       MessageList.QualifiedMessageFileName = InQualifiedMessageFileName;
       MessageList.MessageDataLength        = %Len(InMessageData);
       MessageList.PointerToNextEntry       = *Null;

       //Set the message data pointer to point to end of message list
       //  structure.
       ptrMessageData = ptrMessageList + %Size(MessageList);

       //Now we can move in to message data.
       %Subst(MessageData        :
              1                  :
              %Len(InMessageData)) = InMessageData;

       MessageCount += 1;

       Return;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * GetMessageFromList
     *   Get a message from list.
     *     In     Parms - Pointer to item.
     *     In/Out Parms - None.
     *     Out    Parms - MessageId.
     *                    Qualified Message File Name.
     *                    Message Data.
     *   Return         - True or False. Message Found.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p GetMessageFromList...
     p                 b
     d                 pi                  Like(StdLgl)
     d   InPointerToEntry...
     d                                     Like(StdPtr)
     d                                     Value
     d   OutMessageId...
     d                                     Like(StdMsgId)
     d   OutQualifiedMessageFileName...
     d                                     Like(StdQnam)
     d   OutMessageData...
     d                             1024    Varying
      /Free

       //Make sure we are not getting a null.
       If InPointerToEntry = *Null;
         Return cFalse;
       EndIf;

       //Point to last entry.
       ptrMessageList = InPointerToEntry;

       OutMessageId                = MessageList.MessageId;
       OutQualifiedMessageFileName = MessageList.QualifiedMessageFileName;

       //Get a pointer to Message Data at end of MessageList.
       ptrMessageData = ptrMessageList + %Size(MessageList);

       //Pull out message data from end.
       OutMessageData = %Subst(MessageData                  :
                               1                            :
                               MessageList.MessageDataLength);

       Return cTrue;

      /End-Free
     p                 e
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     * RemoveMessageFromList
     *   Remove a message from list.
     *     In     Parms - Pointer to item.
     *     In/Out Parms - None.
     *     Out    Parms - None.
     *   Return         - None.
     *= = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
     p RemoveMessageFromList...
     p                 b
     d                 pi
     d   InPointerToEntry...
     d                                     Like(StdPtr)
     d                                     Value

     d ptrSaveNext...
     d                 s                   Like(StdPtr)
      /Free

       ptrMessageList = InPointerToEntry;

       ptrSaveNext = MessageList.PointerToNextEntry;

       DeAlloc ptrMessageList;

       If ptrSaveNext = *Null;
         ptrAnchor = *Null;
         ptrTail   = *Null;
       Else;
         ptrAnchor = ptrSaveNext;
       EndIf;

       MessageCount -= 1;

       Return;

      /End-Free
     p                 e
